/************************************/
/*	OS Assignment 3
/*	By: Simon Foucher
/*	260 223 197
/************************************/


/************************************/
/*	includes and definitions				
/************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#define TRUE 1
#define FALSE 0
typedef int BOOL;


/************************************/
/*	Function declarations				
/************************************/
void logo(void);
int interpreter(char commandline[], int Running);


/************************************/
/*	Global Variables & datatypes				
/************************************/
char *FREERAM[100];							//Pointer to RAM array
int FRAMESIZE = 4;                          // Default frame size value
char *shellMemoryVar[100];
char *shellMemoryData[100];
char userPrompt[40] = ".$ ";
BOOL verbose = FALSE;


int quanta = 10;	                            // We'll give the option to the user to set quanta
                                            // Since CPU interrupted by shell prompt every CPU cycle, 
int concurrent = 10;
int readyHead;								// Pointer to the head of the Ready Q
int readyTail;								// Tail pointer of the ready Q 
int readyQ[10];     		                // We'll use an array for the ready queue
int terminateQ[10];
int loadingQ[50];							// Loading Q
int loadingHead;							// Head pointer to the loading Q
int loadingTail;	
					


struct page_rec
{
        int inRAM;							// boolean true if page in RAM, false if page in backing store
        int wasModified; 					// will not be used in this assignment
        int baseAddress; 					// the index number of the RAM array cell of the frame
};

struct PCBdef								// Process Control Block Data Structure definition
	{
		char filename[50]; 					// Stores the name of the script loaded into memory
		int start; 						    // Start address loaded into RAM
		int length; 						// Number of instructions
		int current; 						// Current instruction being executed
		char *memory[20]; 					// List of Shell Memory variables and values
		struct page_rec pageTable[100];		// max limit of 100 pages
		FILE *BACKINGSTORE;					// Backing store file pointer
		long int MAXPAGES;					// total number of created pages for program
		struct PCB *next; 					// Next node in the linked-list
	};


struct cpu
	{
	int quanta; int returnCode;
	int pc;
	char *memory[20];						// Since the PCBs only have 20 memory cells
	};


struct bs_rec								// Backing store
{
        char command1[100];
        char command2[100];
        char command3[100];
        char command4[100];
};






struct PCBdef PCB[10];				// Creates 9 PCBs (max threads that can execute at the same time)



/************************************/
/*	Functions						
/************************************/



/*******************************************************************************/
/*	New/Modified functions since shell.c are clarly identified.
/*	All other functions are similar to old version
/*******************************************************************************/

BOOL displayMemVar(char commandline[])
{
    int i,j;
    BOOL found = FALSE;
    char variable[50];

    // variable names is the second word

    for(i=3,j=0;i<200 && (commandline[i]!='\0' || commandline[i]!='\0');i++)
      if (commandline[i]!=' ')
      {
          variable[j] = commandline[i];
          j++;
      }

    variable[j]='\0';

    // locate variable in shell memory

    for(i=0;i<100 && shellMemoryVar[i]!=NULL;i++)
    {
        if (strcmp(shellMemoryVar[i],variable)==0)
        {
            found = TRUE;
            printf("%s=%s\n",variable,shellMemoryData[i]);
            break;
        }
    }

    if (!found) {printf("Not found\n"); return FALSE;}

    return TRUE;
}

void stripCRLF(char array[])
{
    int i;

    for(i=0;i<500 && array[i]!='\0';i++)
      if (array[i]=='\r' || array[i]=='\n') array[i]='\0';
}


/*******************************************************************************/
/*	Function setMemVar: Stores variable and value
/*	
/*	Input: 
/*      The command (from RAM or SHELL)
/*      Running: a pointer to memory. If Running = 15; command came from SHELL
/*
/*  Additions:
/*      - Records the variable used by a script to its PCB
/*      - The function blocks scripts form overwrighting variables that are
/*          not part of their PCB (either created by shell or another script
/*      - If a shell command overwrights a variable reated by a script,
/*          a warning is issued but the variable get's overwritten
/*******************************************************************************/





setMemVar(char commandline[],int Running)
{
    char variable[100];
    char data[100];
    char userInput[1];
    int i,j,k;
    BOOL outOfMemory = TRUE;
    int script = TRUE;                                          // Assube that we are interpreting a script be default
    int allowed = TRUE;                                         // To make sure that processes don't overwrite other processes's memory          
    if (Running==15)                                            // If the command came from the shell, Running =15                    
        script = FALSE;                                         // Not scripting, use PCB memory

    // skip white spaces
    for(i=3;i<200 && commandline[i]!='\0' && commandline[i]==' ';i++);

    // copy variable
    for (j=0;i<200 && commandline[i]!='\0' && commandline[i]!=' ';i++,j++)
        variable[j]=commandline[i];

    variable[j]='\0';

    // skip white spaces
    for(;i<200 && commandline[i]!='\0' && commandline[i]==' ';i++);

    // copy data
    for (j=0;i<200 && commandline[i]!='\0' && commandline[i]!=' ';i++,j++)
        data[j]=commandline[i];

    data[j]='\0';


    // Store it into the shell's memory, overwritting wit user approval it if it already exists

    for(i=0;i<100;i++)
    {
        if (shellMemoryVar[i]==NULL) // not in memory so insert and exit
        {
            allowed = TRUE;
            shellMemoryVar[i]  = strdup(variable);
            shellMemoryData[i] = strdup(data);
            outOfMemory = FALSE;
            break;
        }

        else if (strcmp(shellMemoryVar[i],variable)==0)                     // If the variable already exists in Shell memory
        {
           if(script)                                                       // If scripting, make sure we don't overwrite other processes's valriables
                {
                    allowed = FALSE;                                        // Not allowd until proven that variable is in script memory
                            for(j=0;(j<=20 && PCB[Running].memory[j] != NULL);j++)// Scanning internal memory for variable name
                            {                                                       
                                if(strcmp(PCB[Running].memory[j], variable)==0)   // If it's part of the script's memory
                                    {allowed = TRUE;break;}                       // Allow overwright and break
                            }

                    if(!allowed)                                            // Warn user of illegal operation
                      {
                        printf("\nWARNING!!!\nThe script is trying to overwright the variable '%s',\n",variable);
                        printf("which was created by another script or the shell.\n\n");
                        printf("Permission denied and instruction ignored!\n\n");
                      }
                }// End of if scripting

            else if(!script)                                                // If we are not scripting
                {
                    allowed = TRUE;                                         // Allowed to overwright                   
                    for(k=0;k<=9;k++)                                       // Scan all PCBs
                      {
                        for(j=0;(j<=20 && PCB[k].memory[j] != NULL);j++)    // Scan every non empty memory Cells of PCBs
                            {
                              if(strcmp(PCB[k].memory[j], variable)==0)     // If the variable is found in a PCB's memory
                                {                                           // Warn user 
                                    printf("WARNING!!!\nThe variable %s was created by another script.\n",variable);
                                    printf("Overwriting it's value will permanently destroy it's previous value\n");
                                    printf("without warning the script who created it. This might cause the process\n");
                                    printf("to behave abnormally or crash.\nAlso note that the variable will");
                                    printf("be deleted from Shell memory\nwhen that script ends!!\n\n");
                                    break;
                                }// End of displaying warning
                            }// End of scanning every non empty memory cells
                      }// End of scanning every PCBs
                }// End of if !scripting

            if(allowed)                                                     // So if we are allowed to overwright
                {                        
                    free(shellMemoryVar[i]);
                    free(shellMemoryData[i]);

                    shellMemoryVar[i]  = strdup(variable);
                    shellMemoryData[i] = strdup(data);

                    outOfMemory = FALSE;
                    break;
                }
        }// End of if variable already exists
    }

                       
    if(outOfMemory == FALSE && script==TRUE && allowed == TRUE)             // If we are scripting and variable got stored in SHELL memory
        {           
            for(j=0;j<20;j++)                                               // Scan PCB memory cells            
                {
                     if (PCB[Running].memory[j]==NULL)                      // Variable not in memory so insert and exit
                        {
                            PCB[Running].memory[j]  = strdup(variable);
                            break;
                        }
                      else if (strcmp(PCB[Running].memory[j],variable)==0)  // Variable already exists in PCB memory, break
                        break;

                }// End of scanning PCB memory cells
        }// End of out of !outOfmemory & scripting

    // if it was not added to the shell the outOfMemory still TRUE

    if (outOfMemory)
        printf("%s=%s not inserted.  Out of shell memory!\n",variable,data);
}



void changePrompt(char commandline[])
{
    int i,j;
    // variable in the second word

    for(i=6,j=0;i<200 && commandline[i]!='\0';i++)
      if (commandline[i]!=' ')
      {
          userPrompt[j] = commandline[i];
          j++;
      }
    stripCRLF(userPrompt);
    userPrompt[j]   = ' ';
    userPrompt[j+1] = '$';
    userPrompt[j+1] = '\0';
	
}


void execute(char commandline[])
{
    int i,j, errorCode=-1;
    char filename[200];
    char scriptcommand[500];
    FILE *ptr;

    // filename in the second word

    for(i=6,j=0;i<200 && commandline[i]!='\0';i++)
      if (commandline[i]!=' ') 
      {
          filename[j] = commandline[i];
          j++;
      }

    filename[j] = '\0';

    // Verify the file is good

    ptr = fopen(filename,"rt");
    if (ptr == NULL)
    {
        printf("Invalid script file\n");
        return;
    }

    // Process the script file to completion

    fgets(scriptcommand,500,ptr);
    stripCRLF(scriptcommand); // more commonly know as CHOMP
//    errorCode = interpreter(scriptcommand);

    while(!feof(ptr) && errorCode != -1)
    {
        fgets(scriptcommand,500,ptr);
        stripCRLF(scriptcommand);
//        errorCode = interpreter(scriptcommand);
    }

    fclose(ptr);
}

/*******************************************************************************/
/*	Function setQuanta: Sets the Quanta for the CPU's Run-time-environment
/*	
/*	Input: 
/*      - qanta(global variable)   1 quanta = 1 instruction per CPU cycle
/*      - commandline[]: user input
/*******************************************************************************/


int setQuanta(char commandline[])
{   // Reading ASCII char as an integer: [ASCII]49 = [int]1 and [ASCII]57 = [int]9
    int j;												// K decides where the character is placed

	quanta = 0;											// Reser to 0 for the algorithm to work
    for (j=6; j<30 && commandline[j]=='\0'; j++)		// Find the 1st non ' ' 

	stripCRLF(commandline); 
// We get kicked out of this loop when commandline[j] = NULL
	j++; 
	while(commandline[j])
		{
			quanta = (quanta*10);					// Shift old value << 1 dec. place
			quanta = quanta + (commandline[j]-48);    // Add LS digit
			j++;
		}

    printf("Quanta has been set to '%d'\n", quanta);
	return quanta;
}


/*******************************************************************************/
/*	Function Pushterminate: Pushes the PID of a PCB into the Terminate Queue
/*      Since we only allow 10 processes to run at the same time, 
/*      the terminate Q can allow 10 members, and it gets cleard every CPU cycles
/*      the list can never overflow
/*
/*	Inputs:
/*		- PID: PID of the process ready for termination 
/*******************************************************************************/
void PushTerminate(PID)
{
    int i;
	for(i=0; i<9;i++)					    // Now scan the Queue from start to end looking for next empty spot
	    {
    		if(terminateQ[i] == -1)         // If no PCB at Queue location 
                {
                    terminateQ[i] = PID;
                    break;
        	     }
        }
}



/*******************************************************************************/
/*	Function PushReadyQueue: Pushes a PCB into the Ready Queue
/*
/*	Inputs:
/*		- readyHead: Pointer to the Head of the Ready queue (declared globally)
/*
/*	Output:
/*		- Pointer to the progream in ready Queue (used as a PID) 
/*******************************************************************************/

int PushReadyQueue(PID)
{
	int i=0;
    if(readyQ[readyHead] == -1)             // Therefore readyHead is pointing to an empty spot
      {
        readyQ[readyHead++] = PID;          // Pushes the PID in the Q
        if(readyHead == 10)                 // Since we only allow up to 10 PCBs, condition implies that
           readyHead = 0;		            // we reached the end of the Queue; need to wrap arround to find space
// For debugging
/*for(i=0; i<10;i++)
	printf("ReadyQ[%d] = %d\n", i, readyQ[i]);

printf("ReadyH = %d, ReadyT = %d\n", readyHead, readyTail);
*/    	return 0;                           // Exit function and return 0 = ok!
      }





		return -1;						    // No empty spot in the list; returns error code
}// End PushReadyQueue




/*******************************************************************************/
/*	Function PopReadyQueue: Pops a PCB from the ready Q
/*
/*	Inputs:
/*		- readyTail: (global) pointer to the tail of the ready Queue
/*		
/*	Output:
/*		- PID of the poped element 
/*******************************************************************************/


int PopReadyQueue()
{
	int popedPID;
	popedPID = readyQ[readyTail];					    // Tail points to the last element of the Q

    if(popedPID != -1)
        {
            readyQ[readyTail++] = -1;                     // Empty spot in the Q
	      
        // Since we implemented a circular list and porgrams are queued on a FIFO basis form the 
        // lowers pointer (0) to the highest (9) then looping back 
	        if(readyTail == 10)
                readyTail = 0;              	       	// Means we reached the end of the Queue => 
        }                                               // need to wrap arround to see if there is space
	  
	return popedPID;	    						    // return Queue pointer as a PID
} // End PopReadyQueue()


/*******************************************************************************/
/*	Function PopLoadingQueue: Pops a PCB from the loading Q
/*	(might get merged with popReadyQ if i have time)
/*	Inputs:
/*		- loadingTail: (global) pointer to the tail of the ready Queue
/*		
/*	Output:
/*		- PID of the poped element 
/*******************************************************************************/


int PopLoadingQ()
{
	int popedPID;
	popedPID = loadingQ[loadingTail];					    // Tail points to the last element of the Q

    if(popedPID != -1)
        {
            loadingQ[loadingTail++] = -1;                     // Empty spot in the Q
        // Since we implemented a circular list and porgrams are queued on a FIFO basis form the 
        // lowers pointer (0) to the highest (9) then looping back 
	        if(loadingTail == 49)
                loadingTail = 0;              	       	// Means we reached the end of the Queue => 
        }                                               // need to wrap arround to see if there is space
	  
	return popedPID;	    						    // return Queue pointer as a PID
} // End PopReadyQueue()

/*******************************************************************************/
/*	Function createPCB: Creates a PCB for 'program'
/*
/*	Inputs:
/*		-RAMptr: "pointer" to start of program in RAM (since RAM is an array; address = int)
/*		-filename: Name of the program
/*		-lenght: number of RAM cells occupied by program (number of commands)
/*
/*	Output: 
/*		PID: not a real PID, since it represents where in the Ready Q the 
/*		PCB is sitting, but good ennough to refer to a process since
/*		only 1 process can be at one PCB at a time
/*
/*  Retunr: PID or -1 if error
/*******************************************************************************/


int createPCB(char filename[], int length)
{	
	int i,k;
	int PID = -1; 

 	// Since we have an array as Q, we need to scan for an 'empty' PCB
	for (i=9; i>=0; i--)
		{
			if(PCB[i].length == 0)
				PID = i;
		}                                  
	
    if(loadingQ[loadingHead] == -1)                             // If there is room in the readyQ
        {
        	for(i=0; i <=50; i++)                   	        // Insert File name; could have used strncpy.
    	     {PCB[PID].filename[i] = filename[i];}		
        	//PCB[PID].start = RAMptr;							// Address in RAM with start of the program	
        	//PCB[PID].current	= RAMptr;						// Since not executing yet, current = start 
        	PCB[PID].length = length;							// Given to us when we looked for empty space
	        memset(PCB[PID].memory, '\0', sizeof(PCB[PID].memory));		// Formats memory
			PCB[PID].current = 0;	 							// Instruction to be executed next
			for(k=0; k<=i; k++)
				{
					PCB[PID].pageTable[k].inRAM = 0; 			// None of the pages are in RAM
					PCB[PID].pageTable[k].wasModified = 0;  	// Cuz i'm sure we'll need it eventually 	
					PCB[PID].pageTable[k].baseAddress = -1; 	// To notify that it never got loaded in RAM
				}
            printf("'%s' has been assigned PID: '0x0%d'\n", filename, PID);
         }
    else
        {
            printf("Too many programs in the loading Q at the moment...\n");
            printf("Try to run this again later!\n\n");
        }

	return PID;
}// End createPCB




/*******************************************************************************/
/*	Function terminateprocesses: Terminates all processes in terminateQ
/*      Since the Q gets cleared every CPU cycle, don't need to worry about 
/*      implementing Queue for fairness
/*
/*	Inputs: terminateQ (Global)
/*
/*	Output: N/A
/*******************************************************************************/

// Close BS, delete BS
void TerminateProcesses()
{
	int i,j, k, t,PID;
	char filenameBS[100];

                                   
    for(t=0; t <= 9; t++)                                 	// Look at every elements in the Q 
    {
    	PID = terminateQ[t];                                // Extract element to be terminates                           
        if(PID != -1)                                       // If we found a valid PID that needs termination 
        {
       		printf("Terminating PID: 0x0%d\n", PID);    	// Implemented while debugging, kept Just for fun...

// Delete shell variables used by process
            for(k=0; (k <=20 && PCB[PID].memory[k] != NULL); k++)           // Scan every non empty memory cells of the PCB
            {              
            	for(j=0;(j<100 && shellMemoryVar[j] != NULL);j++)           // Comapre every SHELL memory variable with the PCB's variables
                {
                	if (strcmp(PCB[PID].memory[k],shellMemoryVar[j])==0)    // When we find a match
                    {
                    	printf("DELETING '%s=%s' from Shell Memory\n",shellMemoryVar[j],shellMemoryData[j]);
                    	free(shellMemoryVar[j]);            // Free the memory
                        shellMemoryVar[j] = '\0';           // And set to NULL
                        free(shellMemoryData[j]);
                        shellMemoryData[j] = '\0';
                        break;
                    }
          		}
          	}

// Delete PCB's record of used variables
   	        memset(PCB[PID].memory, '\0', sizeof(PCB[PID].memory));		// Then format PCB memory

// Free RAM used by process
			for(j=0; j<100; j++)
				if(PCB[PID].pageTable[j].inRAM)
				{
					FREERAM[PCB[PID].pageTable[j].baseAddress] = NULL;     	                    
					PCB[PID].pageTable[j].baseAddress = -1;
					PCB[PID].pageTable[j].inRAM = 0;
				}

// Then delete the backing store file
			strcpy(filenameBS,"BACKINGSTORE/"); 
			k = 13;
		    for(i=0;i<100 && PCB[PID].filename[i]!='\0';i++)				// Loops through i, starting at the first space until we get to '\0'
				filenameBS[i+k] = PCB[PID].filename[i];
			filenameBS[i+k]	  = '.';
			filenameBS[i+k+1] = (PID+48);
			filenameBS[i+k+2] = '.';
			filenameBS[i+k+3] = 'B';
			filenameBS[i+k+4] = 'S';
			filenameBS[i+k+5] = '\0';

			if(fclose(PCB[PID].BACKINGSTORE))
				printf("Error closing backingstore file %s", filenameBS);							

			if(remove(filenameBS))
				printf("Error Deleting backingstore file %s", filenameBS);							

// Then delete the rest of the data from PCB
			for(j=0; j <=20; j++)          
				PCB[PID].filename[j] = '\0';	        		
	        PCB[PID].start = 0;					      
            PCB[PID].current = 0;			 
            PCB[PID].length = 0;
			
          	printf("Done terminating PID: 0x0%d\n", PID);               // Implemented while debugging, kept Just for fun...
			terminateQ[t] = -1;
         }// End if PID = -1
                                           
    }
}



/*******************************************************************************/
/*	Function storeInRAM: 
/*		- Creates Backing store (/BACKINGSTORE/PROG.PCB.BS
/*		- Calls functions to initiate PCB
/*		- Push new PCB in Loading Queue
/*
/*	Inputs: 
/*	   - filename[]: contains the script file's name
/*
/*  Returns:
/*     - PID (redy Queue index) or -1 if not ennough room in RAM to store
/*******************************************************************************/
int storeInRAM(char filename[])
{
    int i,j,k;
    int scriptLenght=0;		
    char scriptcommand[100];
	char filenameBS[100];
    int PID;
    FILE *ptr;						// Script file
	FILE *BS;						// Created Backing Store
	struct PCBdef tempPCB;
	struct bs_rec temp_bs_rec;
  


    if((ptr = fopen(filename,"rt")) == NULL)            // Open the file and check if valid
        {
            printf("Can't open script file\n");         // Notyfy user and return error code -1
            return -1;                                  
        }

    while(!feof(ptr))	            					// Compute the space needed in RAM
    {
        fgets(scriptcommand,100,ptr);                   // Counts the rows filled with text of file
	if(scriptcommand[0] != '\n')scriptLenght++;		    // To avoid counting empty rows
    }
	scriptLenght--;
    printf("Program in '%s' is %d instructions long.\n",filename,scriptLenght);
	if(scriptLenght > 400)
		printf("WANNING!! Page table can only accomodate 400 lines of scritp. Your script got truncated at 400!!\n");

/**************************************************************************************/
/*THE BACKING STORE
/**************************************************************************************/
//    1. Create a PCB attached to a temporary pointer.
	printf("Creating PCB\n");

	PID = createPCB(filename, scriptLenght);			// Creates temporary PCB
	if(PID == -1)                                       // Warns user of cause of error
         return -1;                                     // If PID couldn't be created, Exit function

/**************************************************************************************
   2. fwrite the struct in write mode into a text file having the same name as the script
       file but with the extension .BS (for Backing Store). Normally this file would be
       saved in a hidden area called the backing store, but your backing store will be a
       subdirectory named BACKINGSTORE. You can put this subdirectory in the
       current directory.*/
/**************************************************************************************/

// Directory Backingstore got created in main()
	strcpy(filenameBS,"BACKINGSTORE/"); 
	k = 13;
    for(i=0;i<100 && filename[i]!='\0';i++)				// Loops through i, starting at the first space until we get to '\0'
        {
			filenameBS[i+k] = filename[i];
		}
// TO allow many programs to run simultaneously, the bs files will have the structure:
//   "PROGRAM.PID.BS"   Since every running program gets assigned a unique PID when running,
// Every entry should be unique.
		filenameBS[i+k]	  = '.';
		filenameBS[i+k+1] = (PID+48);
		filenameBS[i+k+2] = '.';
		filenameBS[i+k+3] = 'B';
		filenameBS[i+k+4] = 'S';
		filenameBS[i+k+5] = '\0';

// Look if there already exists an BS entry with same name. Since we are appending when writing, need to delete it
		if(!(remove(filenameBS)))
			{
				printf("WARNING!!\nFound an entry named '%s' in the BS\n",filenameBS);
				printf("Old entry got deleted and will be replaced by this one.\n\n");
			}

		if(!(BS = fopen(filenameBS, "w+")))				// Create the Backing store
			printf("Error creating Backing store\n"); 	
		else
			printf("Creating '%s' for PID0x0%d\n...\n", filenameBS, PID);

/***************************************************************************************
   3. Load the script line-by-line into the page_rec structure.*/
/**************************************************************************************/
	i = 0;												// I will count the pages
// The algorithm could allow unlimited pages, but out page table has a max of 100 entries, so we need to 
// cap the pages at 100

	rewind(ptr);							            // Rewinds ptr to start of the file
    while(!feof(ptr) && j < 100)
	    {												// Extract file commands and store in struct
			j = 0;
// First we extract the comand
			fgets(scriptcommand,100,ptr);
// Skip if its an empty row and read the next one
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
// Copy into struct
			for(k=0;k<100;k++)temp_bs_rec.command1[k]=scriptcommand[k];			

// If we reached EOF, mark J and format command in struct
			if(feof(ptr)){j++;temp_bs_rec.command1[0] = '\0';}


// Repeat the operation 4 times in a row, then start over    
			fgets(scriptcommand,100,ptr);
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
			for(k=0;k<100;k++)temp_bs_rec.command2[k]=scriptcommand[k];			
			if(feof(ptr)){j++;temp_bs_rec.command2[0] = '\0';}

			fgets(scriptcommand,100,ptr);
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
			for(k=0;k<100;k++)temp_bs_rec.command3[k]=scriptcommand[k];			
			if(feof(ptr)){j++;temp_bs_rec.command3[0] = '\0';}

			fgets(scriptcommand,100,ptr);
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
			for(k=0;k<100;k++)temp_bs_rec.command4[k]=scriptcommand[k];			
			if(feof(ptr)){j++;temp_bs_rec.command4[0] = '\0';}
									
/***************************************************************************************
   4. Repeat steps 2 & 3 for all the lines in the script. If the last page did not use all 4
       arrays in the struct, make sure to initialize the unused arrays to '\0'*/
/**************************************************************************************/

			// If we haven't hit 4 EOFs in a row, write the page on the BS
			if(j!=4)
				{
					fwrite(&temp_bs_rec, sizeof temp_bs_rec, 1, BS);
					i++;// Increment page counter
				}

	    }// End while(!eof)
		printf("Total of %d pages created and stored.\n",i);

/**************************************************************************************
   5. When done, create a field in the PCB called FILE *BACKINGSTORE that will
       point to that .BS file. Created another field in the PCB called long int
       MAXPAGES. Save the the total number of created pages for that .BS file.
/**************************************************************************************/
		rewind(BS);
		PCB[PID].BACKINGSTORE = BS;
		PCB[PID].MAXPAGES = i;   // Counter trigered at every page creation

		fclose(ptr);


// Warn user if program is = or longer than Page table size
		if(j==100)printf("Reached maximum of 100 instructions. Anything beyond has been ignored!\n\n");


/**************************************************************************************
   6. Insert the PCB into the Loading queue. (Loading Q implemented using an array[9]
/**************************************************************************************/
	if(loadingQ[loadingHead] == -1)         // Empty spot are marked by -1
      {
        loadingQ[loadingHead] = PID;        // Pushes the PID in the Q
        if(loadingHead == 49)                  // Since we only allow up to 49 PCBs, condition implies that
                loadingHead = 0;              // we reached the end of the Queue; need to wrap arround to find space
        else 
	        loadingHead += 1;                     // Pointer to head of ready Q is incremented s.t. it points to first empty spot
	  }// End if there was room in the Q
	else
		printf("WARNING!\n Too many programs in the loading Q at the moment. Try again later!\n");
	
   	printf("Process initiation completed!\n\n");                                // Implemented while debugging, kept Just for fun...

// PCB is in Loading Queue and will be executed during CPU run-time; 
	return PID;

}// END of storeInRAM function












/*******************************************************************************/
/*	Function run: 
/*      - Initiates the command 'RUN'
/*      - Extracts filename(s) from user input
/*      - calls soreinRAM (described above) for every filenames found
/*	
/*	Inputs: 
/*	   - userInput[]: contains the script file's name
/*******************************************************************************/

int run(char userInput[])
{
	int i, j=0;
	char filename[200];
	int canRun = 0;

// Count if there is still room to run more processes

	for(i=0; i<10; i++)
		if(PCB[i].length == 0)
			{canRun++;
			printf("canRun = %d\n",canRun);
}

// Extract the file name	    
    for(i=4;i<200 && userInput[i]!='\0';i++)	// Loops through i, starting at the first space until we get to '\0'
        {
            filename[j] = userInput[i]; 	    // Extract the filename from the user input
            j++;
            if(userInput[i]==' ')		        // Means there is more than one file name
		        {                               // Need to store the first one in RAM
		            filename[j-1] = '\0';			// Complete the file name
					if(canRun-- > 0)
	          	        storeInRAM(filename);		// Extract script script data, store in RAM and create PCB
					else
						printf("Reached MAX processes the OS can handle\n RUN request for '%s' has been ignored!\n\n", filename);      
			
		            j = 0;				        // Resets j counter for next filename
		        }// End if userinput = ' ' 
	    }// End for loop

// For the last file name of the RUN command; j counter hasen't been reset = file script not processed yet
	if(j!=0)  
	{
	   filename[j] = '\0';			            // Complete the file name
		if(canRun>0)
	       storeInRAM(filename);	    	        // Extract script script data, store in RAM and create PCB
		else
			printf("Reached MAX processes the OS can handle\n RUN request for '%s' has been ignored!\n\n", filename);      
	}    

	return 0;
}// End run();

/*******************************************************************************/
/*	Function: memManager (Memory Manager)
/*	
/*	Input: PID of program (PID.current will be used)
/*	FRAMESIZE: Global variable, can be updates only at launch
/*
/*	What it does:
/*      1. Loads PID.current page  into RAM
/*			if new PCB, loads first page
/*			if called surring run-time, loads next page to be executed
/*      2.A.  Page is loaded using first fit strategy. 
/*		2.B   If all of RAM is full then the manager removes a page from RAM from
/*		the PCB at the tail of the ready list. If that PCB has NO pages in RAM we 
/*		go to the second to last PCB, and so on until we find a PCB with a page in RAM. 
/*			Overwrite the oldest cell, which should be cell 0. If that page is already
/*      not in RAM, we will then overwrite the next oldest which is cell 1, and so on
/*      until we fine one. We only overwrite one page since we only need one page.
/* 			Then updated the victim's page table and overwrite the frame with the new commands.  
/*      3. PCB is added to the tail of the Ready queue.
/*			if new PCB, got removed from loading Q
/*			if running PBC, looses quanta and goes to the back of the readyQ
/*******************************************************************************/

void memManager(int newPID)
{
	int victimePID;
	int RAMptr = -1;
	int current;
	int i,j,k;
	int tempReadyTail = readyTail;


	printf("Starting Memory Manager for PID: 0x0%d\n",newPID);
// Look for an empty FRAME  in RAM

    for(j=0; j<=(100-FRAMESIZE); j++)				    // Scan RAM memory array for empty spot
     {				
		if(FREERAM[j] == '\0')				            // If start of frame = \0, whose frame is empty
		    {	                                           
		        RAMptr = j;	            				// Save location of start of empty FRAME
			    printf("Found an empty FRAME at address RAM0x");
	            if(j < 16)printf("0");                  // Created for debugging, kept just for fun :)
				printf("%x\n",RAMptr);
		        break;                              // Breaks out of for loop s.t. scanningRAM=0 can take effect
	        }// End if found a spot in RAM	
		}//End forJ loop


// If we didn't find a spot in RAM; we need to loook for a victim.
// We push the PCB into the ready Q now, s.t. if it's the only one
// running and hogging the whole RAM array, it will become its own victim.


// First we figure out how many pages we need to extract
	int PPF;										// # Pages per frame 
	PPF = FRAMESIZE/4;								// Since 4 instructions per page (default 1 PPF if FRAMESIZE =4)
	struct bs_rec data;								// In order to lead a frame, we need to extract PPF pages
	 	
// Then look for a victime

							// So we are alone


	for(i=0; (i<10 && (RAMptr == -1)); i++)				// Loop through until we find a spot 			
	{
		victimePID = readyQ[tempReadyTail++];			// Pop the last guy in the readyTail with a temp pointer
		if(tempReadyTail == 10)tempReadyTail = 0;		// Since circular Q

		if(victimePID != -1)
		printf("RAM is full at the moment... Attempting to steal space from PID0x0%d\n",victimePID);

		if(victimePID == -1)
			{
				victimePID = newPID;							// Flag ourselves as a Victime
				printf("RAM is full at the moment... Removing some of PID0x0%d's pages\n",victimePID);
			}

// Now that we found a victime, or ourselves
		printf("Booting off PID:0x0%d's\n",victimePID);
		for(k=0; (k<100 && (RAMptr == -1));k++)			// Scan the victim's page table	
			{
				if(PCB[victimePID].pageTable[k].inRAM)	// Oldest page table in RAM
					{
						RAMptr = PCB[victimePID].pageTable[k].baseAddress;	// Steal his spot in RAM
						for(j=0;j<PPF;j++)							// We knock out one page per pass
						{
							printf("Page #%d out of RAM0x",k);
							if(RAMptr<16)printf("0");
							printf("%x\n",PCB[victimePID].pageTable[k].baseAddress);
							PCB[victimePID].pageTable[k].inRAM = 0;				// Then boot him out!!!
							PCB[victimePID].pageTable[k].baseAddress =0;
							k++;
						}
					}// End if
			}// End for k loop
	}
	printf("\n");
						



// Now RAMptr points to an empty frame         


// Extract the next page(s) to be loaded into RAM the next page


// We'll use 'current' to keep track of which page we are at and update the page table
	current = PCB[newPID].current/4;

// Then we extract the pages and store them remporarly into a data structure
		FILE *fptr = PCB[newPID].BACKINGSTORE;		// The file handle has been left open and rewinded :)

		long offset = current * 400;					// Calculate the offset
		fseek(fptr,offset, 0); 


// We read one page per pass until we reach the last page
		printf("Loading PID:0x0%d's\n",newPID);
		for(k=0; (k<PPF && current < PCB[newPID].MAXPAGES);k++)							
			{
				if(feof(fptr) )    // check for EOF
				{
					printf("Reached the end of the BS for PID0x0%d\n",newPID);
			    	break;
				}
				if(!(fread(&data, sizeof data, 1, fptr)))
					printf("Error reading from Backing Store.\n");
// Load page(s) in RAM starting at address RAMptr and update page table

			printf("Page #%d into RAM0x", current);			
			if(RAMptr<16)printf("0");
			printf("%x\n", RAMptr);
			   PCB[newPID].pageTable[current].inRAM = 1;
			   PCB[newPID].pageTable[current].baseAddress = RAMptr; 
               FREERAM[RAMptr] = strdup(data.command1);
				RAMptr++;
               FREERAM[RAMptr] = strdup(data.command2); 
				RAMptr++;
               FREERAM[RAMptr] = strdup(data.command3);
				RAMptr++;
               FREERAM[RAMptr] = strdup(data.command4); 
     			current++;
				RAMptr++;
// FOR DEBUGGING  
				// printf("k=%d\n, Command1: %s,Command2: %s,\nCommand3: %s, Command4: %s \n",k, data.command1,data.command2,data.command3,data.command4);
			}
			printf("\n");

//			rewind(fptr);


// FOR DEBUGGING  
//for(k=0;k<100;k++)
//	printf("RAM[%d] = %s\n", k, FREERAM[k]);

// The process is pushed into the ready Q externally

	
     
}






/*******************************************************************************/
/*	Function: Command line interpreter
/*	
/*	Changed since shell.h:
/*		RUN: ENNABLE virtual CPU Running of multiple 'programs'
/*		ECHO: Echos on the screen what follows
/*		EXIT: definition was extended to (EXIT or LOGOUT) 
/*******************************************************************************/

int interpreter(char userInput[], int Running)
{
    char response;
    int i, j;
    char extraChars[10];
    if (verbose)
    {
        printf("Verbose mode: Confirm execution of %s. (y/n): ",userInput);
        scanf("%c",&response);
        fgets(extraChars,10,stdin);
	stripCRLF(extraChars);
        if (response == 'N' || response == 'n')
        {
            printf("Operation cancelled\n");
            return 0;
        }
    }

    if (strcmp(userInput,"EXIT")==0 || strcmp(userInput,"LOGOUT")==0)
        return -1;
    else if (strcmp(userInput,"CLS")==0)
    {
        for (i=0;i<30;i++) printf("\n");
    }
    else if (strcmp(userInput,"HELP")==0)
    {
        printf("\n\nmyshell Command Set\n\n");
        printf("EXIT            to terminate the shell\n");
        printf("CLS             to clear the screen\n");
        printf("HELP            displays this help screen\n");
        printf("VERBOSE ON/OFF  to toggle verbose mode\n");
        printf("SET VAR VALUE   add a variable to shell memory\n");
        printf("GET VAR         see the value of a shell memory var\n");
        printf("SCRIPT FILENAME execute a script\n");
	printf("QUANTA [int]	sets quanta time to [int] (default = 3)\n");
        printf("PROMPT VALUE    change your prompt\n");
        printf("\n");
    }
    else if (strcmp(userInput,"VERBOSE ON")==0)
        verbose = TRUE;
    else if (strcmp(userInput,"VERBOSE OFF")==0)
        verbose = FALSE;
    else if (strcmp(userInput,"GET")==0)
    {
        printf("Shell Memory Variables\n");
        for(i=0;i<100 && shellMemoryVar[i]!=NULL;i++)
        {
            printf("%s=%s\n",shellMemoryVar[i],shellMemoryData[i]);
        }
    }
    else if (strncmp(userInput,"GET",3)==0)
        displayMemVar(userInput);
    else if (strncmp(userInput,"SET",3)==0)
        setMemVar(userInput, Running);
    else if (strncmp(userInput,"PROMPT",6)==0)
        changePrompt(userInput);
    else if (strncmp(userInput,"SCRIPT",6)==0)
        execute(userInput);
    else if (strncmp(userInput,"RUN",3)==0)
	    run(userInput);
    else if (strncmp(userInput,"ECHO",4)==0)
	{
		for(i=5;i<200 && userInput[i]!='\0';i++)
			{putchar(userInput[i]);}
		printf("\n");
	}
   else if (strncmp(userInput,"QUANTA",6)==0)
	setQuanta(userInput);
    else
        system(userInput);

    return 0;
}



/*******************************************************************************/
/*	Function: shell()
/*	
/*	The main command-line control loop
/*******************************************************************************/


int shell()
{
    int i, errorResult=0;
    char response;
    char userInput[200];
    int Running = 15;               // Tells the interpreter that the shell is sending the command 

    // Display the command-line prompt
    printf("%s", userPrompt);
    fgets(userInput,200,stdin);
	stripCRLF(userInput);
    errorResult = interpreter(userInput, Running);

    return errorResult;
}



/****************************************************/
/*	Function: switches()
/*
/*	inputs:
/*		argc 	= argv[] element count
/*		argv[] 	= contains all the arguments inputed
/*			argv[0]	= 	launchcpde
/*			argv[1] = 	parameter1
/*			argv[n] = 	parametern
/*			argv[n+1] =	null
/*
/*	outputs:
/*		Returns 0 if no inputs => Programm launches
/*		Returns 1 if an argument was inputed => Program terminates
/*		Returns -1 if illegal argument input => Program terminates
/*
/*	Function:
/*		Acts on decoded argument
/*
/*
/****************************************************/
int switches(int argc, char **argv)
{
	if(argc == 0)return 0;						//No launch option, function returns 0 to main
    char framesize[3];
	int i, j, k;							

	for( i = 1; i <= argc; i++ )				//Scan through arguments; argv[argc]=null
     {
       switch((int)argv[i][0] )					//Check for '-'
       {
       case '-':					    		//Found '-'; interprets the option
			switch((int)argv[i][1])				//scans the 1st letter of arguments after '-'
                {
                   case 'h':					//Found the option -h; displaying help
					   puts("\nOption '-h' => Displaying Launch Parameters\n"
					        "To launch the OS, input 'os' -[argument] from the command line.\n"
							"Type 'os -h' to display help file\n"
							"The operating system was NOT launched...\n\n");
							return 1;       	//Tells main that valid option was found
                            break;				//Not nescessary.
                    case 'f':					// Possible framesize
                        if (strncmp(argv[i], "-framesize",10) == 0)
                            { 
                            k = 0;												// K decides where the character is placed
							FRAMESIZE = 0;										// Reser to 0 for the algorithm to work
                            for (j=10; j<30 && argv[i][j]=='\0'; j++)			// Find the 1st non ' ' 
                            {}  
							while(argv[i][j])
								{
									if(k)										// Means that this is not the 1st time evaluating
										FRAMESIZE = (FRAMESIZE*10);				// Shift old value << 1 dec. place
			
									FRAMESIZE = FRAMESIZE + (argv[i][j]-48);    // Add LS digit
									k++;										// Notify that we tampered with FRAMESIZE in case of next pass

									j++;   
								}
							concurrent = 100/FRAMESIZE;
							if(FRAMESIZE > 100)
								{
									printf("\nWARNING!!! Only 100 slots in RAM!!\n");
									printf("To avoid instabilities, framesize cannot exceed 100!\n"); 
									FRAMESIZE = 100;
								} 

							if(FRAMESIZE%4)
								{
									printf("\nWARNING!!! FRAMESIZE must be a multiple of 4!\n");
									printf("Regressing to an adequate value\n"); 
									FRAMESIZE -= FRAMESIZE%4;

								}                          
							printf("FRAMESIZE has been set to: %d\n\n", FRAMESIZE);

								
                            }// End switch
                            return 0;
                            break;
                   default:						//If no valid launch parameter have been found
				 		printf( "-%c is not a valid launch parameter\n", argv[i][1]);
                        return -1;				//Informs main that user input illegal launch parameter
                        break;
                 }// End of scanning -[parameter]

			break;// End of case '-'; no legal parameter calls have been found

       default:		    						//If parameter doesn't start with '-'; illegal
	   		puts("Illegal syntax: launch parameter must start with '-'\n"
				 "OS was not launched...\n\n");
			return -1;
            break;
		}// End of switch loop scanning parameters
      }// End of for loop scanning parameters
return -1;				    				    //The program shouldn't be able to get here
}//End of function switches()







/*******************************************************************************/
/*	Function: TaskSwitchIn(Running, CPUptr)
/*
/*
/*	Inputs:
/*		- Running: 'pointer' (int) to the PBC that got poped from ready Q and gets to run
/*		- CPUptr: real pointer to the CPU => defined as a strucu in main()
/*
/*	Outputs:
/*		None
/*
/*	Function: Transfer the information from PCB to CPU
/*	Returns: None
/*
/*	*remember Datatypes definitions:
/*     PCB(char filename[50], int start, int length, int current, char *memory[20], struct PCB *next)
/*	CPU(int quanta, int returnCode,	int pc,	char *memory[20])
/*******************************************************************************/

void TaskSwitchIn(struct cpu *CPUptr, int Running)
{
	int i;	
	CPUptr->quanta = quanta;					// Reset quanta before starting job
	CPUptr->returnCode = 0;						// Reset CPU return code
	CPUptr->pc = PCB[Running].current;      	// Set the PC to where it left off
}


/*******************************************************************************/
/*	Function: TaskSwitchOUT(Running, CPUptr)
/*
/*
/*	Inputs:
/*		- Running: 'pointer' (int) to the PBC that got poped from ready Q and gets to run
/*		- CPUptr: pointer to the CPU => defined as a strucu in main()
/*
/*	Outputs:
/*		None
/*
/*	Function: Transfer the information from CPU to PBC
/*	Returns: None
/*
/*	*remember Datatypes definitions:
/*     PCB(char filename[50], int start, int length, int current, char *memory[20], struct PCB *next)
/*	CPU(int quanta, int returnCode,	int pc,	char *memory[20])
/*******************************************************************************/

void TaskSwitchOut(struct cpu *CPUptr, int Running)
{
    PCB[Running].current = CPUptr->pc;                 	// Records PC
}





/****************************************************/
/*	Function: formatRAM()
/*
/*	inputs:
/*		char *FREERAM[100] (A global variable
/*
/*	outputs:
/*		NONE
/*
/*	Function:
/*		Initializes FREERAM array by formating every
/*		cells (sets every elements to NULL)
/*
/*	Returns:
/*		0 if everything went well
/*		1 if an error occured
/*
/****************************************************/
int formatRAM(void)
{
	puts("Formating RAM...");
	memset(FREERAM, '\0', sizeof(FREERAM));
	puts(" ...Done!\n\n");
	return 0;
}// End of function formatRAM()




/*******************************************************************************/
/*	Function: login
/*
/*	Inputs: 
/*      - pass.txt: External password file conatining 
/*          NAME1
/*          PASSWORD1
/*          NAME1
/*          PASSWORD2
/*          [...]
/*          END
/*
/*	Function:
/*		Prompts for a name and password whitch are matched with an encrypted text file
/*		containing all valid credentials
/*
/*	Returns:
/*		0 Invalid Name or Password
/*		1 Valid login credentials
/*
/*******************************************************************************/
int login(void)
{
	FILE *fp;
	int i = 0;	
	int havename;
	int havepassword;
	
    struct credentials
	{
		char name[15];
		char pass[15];
	}user, fromfile;
	memset(user.name, '\0', sizeof(user.name));		// Formated to avoid collision login with existing memory values
	memset(user.pass, '\0', sizeof(user.pass));		// Formated to avoid collision login with existing memory values

	logo();									        // Displays OS logo... functionally useless

	printf("\nLogin name: ");						// Name is restricted to 15 characters to avoid iostream buffer overflow
	fgets(user.name, 15, stdin);                    // Get user input name

	printf("Password: ");                           // Get user input password
	fgets(user.pass, 15, stdin);					// Password also restricted to 15 characters

	if((fp = fopen("pass.txt","r"))==NULL)		    //Open the password file
		{
			puts("Cannot open password file!\n"		
				 "Please make sure you have a file named 'pass.txt'\n"
				 "containing login credentials in this directory.\n");
			return 0;
		}

/********************/
/*	Authentication	
/********************/


 	while (feof(fp) == 0)
  	{
		memset(fromfile.name, '\0', sizeof(fromfile.name));		// Format filename buffer avoid accidental
                                                                // login with existing memory values
    	fgets(fromfile.name, 15, fp);		        			// Read next record on pass.txt file
		havename = 1;	                						// Assume a valid user name until proven wrong
		for(i = 0; i < sizeof(user.name); i++)	        		// Check if valid user name
    		{
    			if(fromfile.name[i] != user.name[i])	        // Character by character analysis
    				{
                        havename=0;			                    // Any divergence between the 2 values 
						break;                                  // kicks the porogram out of this loop
                    }// End of looking for invalid name
		    }// End for(scanning name)

		if(havename)		                					// Still have valid name after matching with file values
			{							                        // We have to look for valid password
			memset(fromfile.pass, '\0', sizeof(fromfile.pass));	// Start by formating the password buffer
			fgets(fromfile.pass, 15, fp);	        			// Extract password string associated with name
			havepassword = 1;
			if(fclose(fp))puts("File close error.\n");	    	// Since user names unique, pass file is no longer needed

			while(havepassword)
				{
					for(i = 0; i < sizeof(user.pass); i++)	    // Scan for divergence between record and input passwords
						{
							if(fromfile.pass[i] != user.pass[i])// If != detected, invalid credentials = return 0
								{
									puts("\nInvalid password!\n\n");
									return 0;	                // Exit the function
								}
						}				
					puts("Loging in... ...\n");                 // Otherwise
					return 1;				                    // Code can only get here if divergence not detected
				}// End of havepassword
			}// End of havename; wrong password; function should have already terminated
	}// Reached EOF without finding a valid user name

	puts("\nInvalid user name\n\n");
	if(fclose(fp))puts("File close error.\n");	    			//Close the password file
	return 0;								                    //Return 0 by default
}//End of login script










int main (int argc, char *argv[])
{

	int returnCode = 1;
    int progEnd, i;
    char instruction[100];
	int PIDloading;
	int RAMAddress;	
	int DONE;
	int pageNumber;
	int numRunning;
	int numWaiting;

/****************Step 1: Check for switches***********************************/

	if(--argc)
		if(switches(argc, argv))
			return -1;
	printf("Because of the FRAMESIZE\n", FRAMESIZE);
	printf("Only %d script(s) will be able to run simultaneously\n",concurrent);
	printf("(Running %d scripts will greatly decrease the performance of the machine!)\n\n",concurrent);
	

/****************Step 2: Initialize RAM***************************************/
	if(formatRAM())puts("WARNING!\n\nThe OS detected an error while formating RAM");

/****************Step 3: Initialize Run-Time Environment************************/

    for(i=0; i<10; i++)                             // Initialization of Queues
        {
            readyQ[i] = -1;                         // -1 is used to indicate an empty spot (since all PIDs are >= 0
            terminateQ[i] = -1;
		}
	for(i=0; i<50; i++)                             // Initialization of Loading Q
		loadingQ[i] = -1;

	readyHead = 0;							        // Initialization of Q pointers
    readyTail = 0;  
	loadingHead = 0;							    // Initialization of Q pointers
    loadingTail = 0;                                               
	int PIDRunning;								    // Variable to Identify Running PCB
/*******************************************************************************/
/*	3. Initialization of Virtual CPU with a DATA structure
/*******************************************************************************/

	struct cpu CPU;
	struct cpu *CPUptr;
	CPUptr = &CPU;	

	CPU.quanta = quanta; 							    // The value of quanta can be set by user
	CPU.returnCode = 0;
	CPU.pc = 0;
	memset(CPU.memory, '\0', sizeof(CPU.memory));


/****************Step 4: Run-Time Environment**********************************/


	mkdir("BACKINGSTORE", 0777);// Cuz we'll need it


while(!login());						            	        //Can only pass this step if valid credentials
   {
   printf("Loading shell\n\n");

   while (returnCode != -1)
     {  
        printf("\n");                                           // Just for esthetics
	    returnCode = shell();	 					            // returns -1 when user inputs EXIT or LOGOUT
// For debugging Qs:
//for(i=0;i<10;i++)
//printf("Loading[%d] = %d   Ready[%d] = %d, terminate[%d] = %d\n",i,loadingQ[i],i,readyQ[i],i,terminateQ[i]);}
//printf("Ready H=%d T=%d \nLoading H=%d T=%d\n",readyHead,readyTail,loadingHead, loadingTail);
	

// Figure out who gets to run next   	    
		PIDRunning = PopReadyQueue();                           // Pops next PCB from ready Q
// If there is no one to run, reset Q pointers

// Run next guy on the running Q
        if(PIDRunning != -1)
            {
                printf("CPU is running PID 0x0%d (called '%s')\n", PIDRunning, PCB[PIDRunning].filename); 
	            TaskSwitchIn(CPUptr, PIDRunning);	    		// Sets PC, pops variable from 'stack' to 'CPU registers'
	            CPU.quanta = quanta; 					        // Recursive step because this is done in TaskSwitchIn()...
				DONE = 0;										// Not done

                while(CPU.quanta > 0)                           // while Quanta left	       
              	   {

// If the instruction is in RAM
					pageNumber = CPU.pc/4;												// First compute which page contains the instruction
					if((PCB[PIDRunning].pageTable[pageNumber].inRAM))					// Then look up the page table
					{
						RAMAddress = PCB[PIDRunning].pageTable[pageNumber].baseAddress;	// Look up base address for the page
						RAMAddress += CPU.pc%4;											// Calculate instruction address offset

// FOR DEBUGGING
// printf("PC = %d  Page#= %d RAM address=%d\n",CPU.pc,pageNumber,RAMAddress);


// Execute the instruction
                      	CPU.returnCode = interpreter(FREERAM[RAMAddress], PIDRunning);  // Send command and PCB memory address		
              		  	CPU.pc++;						                        		// Increment CPU program counter		
                    	CPU.quanta --;
					 }// End of if the instruction was in RAM


// Else, if the current instruction of the PCB[PIDrunning] is not in RAM, call the memory manager to store it there
					else
					{
						printf("Instruction %d is not in RAM...\nFetching it in the BS and putting PID0x0%d to sleep.\n\n", CPU.pc, PIDRunning);
						PCB[PIDRunning].current = CPU.pc;		// Since the memMGR relies on this info to build page table
						memManager(PIDRunning);					// Page reference for next frame to be loaded in RAM
						CPU.quanta = 0;							// Kicks us out of CPU until next time
					}

// Then, evaluate weather an EXIT command got processes, or we are at the end of the script
                     if(CPU.returnCode == -1 || CPU.pc == PCB[PIDRunning].length)       // Condition that program is done running
	               	   {    
                     	printf("Done executing or reached EXIT/LOGOUT command.\n\n", PIDRunning);
                    	CPU.quanta = 0;				       		                        // Break out of CPU quanta loop
						DONE = 1;														// Records the fact that we are done executing                    	
						}

                	
					}// End of CPU Quanta

// If the program was done executing
				if(DONE)
                	PushTerminate(PIDRunning);	       				                // Pushes the program into the terminate Queue					

// Otherwise, if there was still some instructions to execute
				if(!DONE)
					{
						TaskSwitchOut(CPUptr, PIDRunning);                  // Save runtinme environment in PCB 
						PushReadyQueue(PIDRunning);                     // Pushes current program into the Ready Q
					}
            }// End if(PIDRunning != -1)

// More OS Overhead
        TerminateProcesses();

// Count how many processes are running
	numRunning = 0;
	for(i=0; i<10;i++)
		if(readyQ[i] != -1)									
			numRunning++;

// Count how many processes are waiting to load (just for fun)
	numWaiting = 0;
	for(i=0; i<50;i++)
		if(loadingQ[i] != -1)									
			numWaiting++;
	
// If there are less processes running that what is allowed based on framesize
	if(numRunning < concurrent)
	{
		PIDloading = PopLoadingQ();								// Pop the loading Q to see if we need to load a program

// Some OS Overhead: Load next guy on the loading Q (if any)
		if(PIDloading != -1)
			{
				memManager(PIDloading);				// Loads PCB's first page int RAM, then pushes him into readyQ by the memMGR
				PushReadyQueue(PIDloading);                     // Pushes current program into the Ready Q
			}
	}

// For fun
		printf("There is/are %d Processe(s) running at the moment\n(%d Allowed in total, %d Waiting)\n",numRunning,concurrent, numWaiting);


    }// End while LOGGED into the SHELL
	puts("Logging out of the shell and exiting OS\n");

}//End of while loop initiating CPU environment... Equivalent of closing computer
// The code can't get here.
return 0;
}// End of main()


// Just for fun, because programing gets boring after a while...
void logo(void)
{
puts("\n"
"           (((\n"
"           {  '>\n"
"           / ,)\n"
"          /_/ /\n"
"==========//=n===========\n"
"         /\n\n"
"    WELCOME TO BIRDX\n\n");
}

